#include <unistd.h>
#include <sys/wait.h>
#include <pthread.h>
#include <fcntl.h>
#include <semaphore.h>
#include "common.h"

#define M 4 // processes number
#define N 2 // thread numbers
#define SEMAPHORE_NAME      "/shared_sem"
#define NUM_RESOURCES 2

typedef struct thread_args_s {
    int child_id;
    int thread_id;
	sem_t* thread_sem;
	int*   shared_thread_var;
} thread_args_t;

sem_t* process_sem;

void *thread_routine(void *args) {
	
	int ret = 0;
    thread_args_t *thread_args = (thread_args_t*)args;
	printf("[CHILD%d] thread %d avviato\n", thread_args->child_id, thread_args->thread_id);
	
	/** 
	 *	gestione sezione critica inter-thread
	 **/
	ret = sem_wait(thread_args->thread_sem);
	ERROR_HELPER(ret, "Errore nella sem_wait");

	// sezione critica inter-thread: scrittura su variabile condivisa 'shared_thread_var'
	*(thread_args->shared_thread_var) += thread_args->child_id * thread_args->thread_id;
		
	ret = sem_post(thread_args->thread_sem);
	ERROR_HELPER(ret, "Errore nella sem_post");
	
	printf("[CHILD%d] thread %d; valore = %d\n", thread_args->child_id, thread_args->thread_id,
												*(thread_args->shared_thread_var));
	/** fine gestione sezione critica inter-processo **/
								
    pthread_exit(NULL);
}

void child_routine(int child_id) {
	printf("[CHILD%d] figlio avviato\n", child_id);
	
	int i = 0, ret = 0;
	
	int* shared_thread_var = malloc(sizeof(int));
	sem_t* thread_sem = malloc(sizeof(sem_t));
	
	ret = sem_init(thread_sem, 0, 1);
	ERROR_HELPER(ret, "Errore inizializzazione thread_sem");
	
	*shared_thread_var = 0;
	
	pthread_t threads[N];
    thread_args_t threads_args[N];
	
	// ciclo di creazione degli N thread
	for(; i < N; i++) {
		threads_args[i].child_id = child_id; 
		threads_args[i].thread_id = i+1; // 'i+1' e non 'i' per far partire thread_id da 1
		threads_args[i].thread_sem = thread_sem;
		threads_args[i].shared_thread_var = shared_thread_var;
		
		ret = pthread_create(&threads[i], NULL, thread_routine, &threads_args[i]);
		PTHREAD_ERROR_HELPER(ret, "Errore creazione producer thread");
	}
	
	// ciclo di join per attendere gli N thread
	for (i = 0; i < N; i++) {
        ret = pthread_join(threads[i], NULL);
        PTHREAD_ERROR_HELPER(ret, "Errore join thread");
    }
	printf("[CHILD%d] esecuzione thread terminata. valore=%d\n", child_id, *shared_thread_var);
	
	/** 
	 *	gestione concorrenza inter-processo
	 **/
	ret = sem_wait(process_sem);
	ERROR_HELPER(ret, "Errore nella sem_wait di process_sem");
	
	// sezione critica inter-processo: simuliamo l'accesso ad una risorsa condivisa tra processi
	printf("[CHILD%d] RISORSA ACQUISITA! Valore = %d\n", child_id, *shared_thread_var);
	
	ret = sem_post(process_sem);
	ERROR_HELPER(ret, "Errore nella sem_post di process_sem.");
	/** fine gestione concorrenza inter-processo **/
	
	// chiusure semafori e rilascio risore allocate dinamicamente
	ret = sem_destroy(thread_sem);
    ERROR_HELPER(ret, "Errore destroy thread_sem");
	
	ret = sem_close(process_sem);
	ERROR_HELPER(ret, "Errore chiusure process_sem");
	
	free(shared_thread_var);
	free(thread_sem);
	
	/* NOTA: qui il semaforo 'process_sem' va chiuso ma non unlinkato,
	 * altrimenti nessun altro processo potrà usarlo
	 */
}

int main() {

	int i = 0, ret = 0;
	
	// apertura semaforo named; in caso di fallimento riprovo dopo l'unlink
	process_sem = sem_open(SEMAPHORE_NAME, O_CREAT | O_EXCL, 0600, NUM_RESOURCES);
	if (process_sem == SEM_FAILED && errno == EEXIST) {
        printf("[WARNING] Il semaforo named esiste già.\n");	
		sem_unlink(SEMAPHORE_NAME);
		process_sem = sem_open(SEMAPHORE_NAME, O_CREAT | O_EXCL, 0600, NUM_RESOURCES);
    }
    if (process_sem == SEM_FAILED) ERROR_HELPER(-1, "[FATAL ERROR] Impossibile aprire il semaforo named");
	printf("semaforo creato, avvio i figli\n");
	
	// ciclo di creazione degli M processi figlio
	pid_t pid;
	for (; i < M; i++) {
		printf("ciclo: %d-esima iterazione\n", i);
		pid = fork();
		if (pid == -1) ERROR_HELPER(pid, "Errore nella creazione del figlio");
		if (pid == 0) {
			child_routine(i+1);
			break; // IMPORTANTE: senza questa break i figli continuano a creare altri figli!!!
		}
		// NOTA: non è necessario indicare un ramo else per il padre poichè lui deve solo creare i figli
	}	
	
	// condizione di verifica se sono il processo padre o il figlio
	if (i == M) { // sono il padre poichè il contatore 'i' è arrivato alla fine del ciclo
	
		// attendo termine degli M figli
		for (i = 0; i < M; i++) {
			ret = wait(NULL);
			ERROR_HELPER(ret, "Errore nella wait di un figlio");
		}
	
		ret = sem_close(process_sem);
		ERROR_HELPER(ret, "Errore nella chiusura del semaforo named");
		
		/* unlink semaforo named: potevo farlo prima? No, solo perchè qui ho la certezza che
		 * il semaforo non serva più poiché sono terminati tutti i figli.
		 */
		sem_unlink(SEMAPHORE_NAME);
		printf("processo padre terminato con successo.\n");
	}
	else { // sono il figlio 'i'
		printf("processo figlio terminato con successo\n");
	}	
	
	return EXIT_SUCCESS;
}	